@@ -23,6 +23,8 @@
 #include "config.h"
 #endif
 
+#define GDK_PIXBUF_ENABLE_BACKEND
+
 #include "eog-image.h"
 #include "eog-image-private.h"
 #include "eog-debug.h"
@@ -46,6 +48,7 @@
 #include <glib/gi18n.h>
 #include <gtk/gtk.h>
 #include <gdk-pixbuf/gdk-pixbuf.h>
+#include <gdk-pixbuf/gdk-pixbuf-io.h>
 #include <libgnomevfs/gnome-vfs.h>
 #include <libgnomeui/gnome-thumbnail.h>
 
@@ -167,6 +170,11 @@ eog_image_dispose (GObject *object)
 		priv->trans = NULL;
 	}
 
+	if (priv->trans_autorotate) {
+		g_object_unref (priv->trans_autorotate);
+		priv->trans_autorotate = NULL;
+	}
+
 	if (priv->undo_stack) {
 		g_slist_foreach (priv->undo_stack, (GFunc) g_object_unref, NULL);
 		g_slist_free (priv->undo_stack);
@@ -240,6 +248,7 @@ eog_image_init (EogImage *img)
 	img->priv->status = EOG_IMAGE_STATUS_UNKNOWN;
 	img->priv->undo_stack = NULL;
 	img->priv->trans = NULL;
+	img->priv->trans_autorotate = NULL;
 	img->priv->data_ref_count = 0;
 #ifdef HAVE_EXIF
 	img->priv->orientation = 0;
@@ -309,7 +318,7 @@ eog_image_update_exif_data (EogImage *image)
 	bo = exif_data_get_byte_order (priv->exif);
 
 	/* Update image width */	
-	entry = exif_content_get_entry (priv->exif->ifd [EXIF_IFD_EXIF], EXIF_TAG_PIXEL_X_DIMENSION);
+	entry = exif_data_get_entry (priv->exif, EXIF_TAG_PIXEL_X_DIMENSION);
 	if (entry != NULL && (priv->width >= 0)) {
 		if (entry->format == EXIF_FORMAT_LONG)
 			exif_set_long (entry->data, bo, priv->width);
@@ -318,9 +327,9 @@ eog_image_update_exif_data (EogImage *image)
 		else
 			g_warning ("Exif entry has unsupported size");
 	}
-	
+
 	/* Update image height */	
-	entry = exif_content_get_entry (priv->exif->ifd [EXIF_IFD_EXIF], EXIF_TAG_PIXEL_Y_DIMENSION);
+	entry = exif_data_get_entry (priv->exif, EXIF_TAG_PIXEL_Y_DIMENSION);
 	if (entry != NULL && (priv->height >= 0)) {
 		if (entry->format == EXIF_FORMAT_LONG)
 			exif_set_long (entry->data, bo, priv->height);
@@ -329,6 +338,19 @@ eog_image_update_exif_data (EogImage *image)
 		else
 			g_warning ("Exif entry has unsupported size");
 	}
+	
+	/* Update image orientation */	
+	entry = exif_data_get_entry (priv->exif, EXIF_TAG_ORIENTATION);
+	if (entry != NULL) {
+		if (entry->format == EXIF_FORMAT_LONG)
+			exif_set_long (entry->data, bo, 1);
+		else if (entry->format == EXIF_FORMAT_SHORT)
+			exif_set_short (entry->data, bo, 1);
+		else
+			g_warning ("Exif entry has unsupported size");
+
+		priv->orientation = 1;
+	}
 #endif
 }
 
@@ -392,6 +414,31 @@ eog_image_real_transform (EogImage     *img,
 	}
 }
 
+static void
+eog_image_pre_size_prepared (GdkPixbufLoader *loader,
+			     gint width,
+			     gint height,
+			     gpointer data)
+{
+	EogImage *img;
+	GdkPixbufFormat *format;
+
+	eog_debug (DEBUG_IMAGE_LOAD);
+
+	g_return_if_fail (EOG_IS_IMAGE (data));
+
+	img = EOG_IMAGE (data);
+
+	format = gdk_pixbuf_loader_get_format (loader);
+
+	if (format) {
+		/* FIXME: We should not be accessing this struct internals
+ 		 * directly. Keep track of bug #469209 to fix that. */
+		img->priv->threadsafe_format =
+			(format->flags & GDK_PIXBUF_FORMAT_THREADSAFE);
+	}
+}
+
 static void
 eog_image_size_prepared (GdkPixbufLoader *loader, 
 			 gint             width, 
@@ -414,7 +461,9 @@ eog_image_size_prepared (GdkPixbufLoader *loader,
 	g_mutex_unlock (img->priv->status_mutex);
 
 #ifdef HAVE_EXIF
-	if (!img->priv->autorotate)
+	if (img->priv->threadsafe_format && (!img->priv->autorotate || img->priv->exif))
+#else
+	if (img->priv->threadsafe_format)
 #endif
 		g_signal_emit (img, signals[SIGNAL_SIZE_PREPARED], 0, width, height);
 }
@@ -442,20 +491,21 @@ eog_image_needs_transformation (EogImage *img)
 {
 	g_return_val_if_fail (EOG_IS_IMAGE (img), FALSE);
 
-	return (img->priv->trans != NULL);
+	return (img->priv->trans != NULL || img->priv->trans_autorotate != NULL);
 }
 
 static gboolean
 eog_image_apply_transformations (EogImage *img, GError **error)
 {
 	GdkPixbuf *transformed = NULL;
+	EogTransform *composition = NULL;
 	EogImagePrivate *priv;
 
 	g_return_val_if_fail (EOG_IS_IMAGE (img), FALSE);
 	
 	priv = img->priv;
 
-	if (priv->trans == NULL) {
+	if (priv->trans == NULL && priv->trans_autorotate == NULL) {
 		return TRUE;
 	}
 
@@ -468,8 +518,17 @@ eog_image_apply_transformations (EogImage *img, GError **error)
 		return FALSE;
 	}
 
-	if (priv->trans != NULL) {
-		transformed = eog_transform_apply (priv->trans, priv->image, NULL);
+	if (priv->trans != NULL && priv->trans_autorotate != NULL) {
+		composition = eog_transform_compose (priv->trans, 
+						     priv->trans_autorotate);
+	} else if (priv->trans != NULL) {
+		composition = g_object_ref (priv->trans);
+	} else if (priv->trans_autorotate != NULL) {
+		composition = g_object_ref (priv->trans_autorotate);
+	}
+
+	if (composition != NULL) {
+		transformed = eog_transform_apply (composition, priv->image, NULL);
 	}
 
 	g_object_unref (priv->image);
@@ -485,14 +544,18 @@ eog_image_apply_transformations (EogImage *img, GError **error)
 			     _("Transformation failed."));
  	}
 
+	g_object_unref (composition);
+
 	return (transformed != NULL);
 }
 
-static GnomeVFSFileSize
-eog_image_determine_file_bytes (EogImage *img, GError **error)
+static void 
+eog_image_get_file_info (EogImage *img, 
+			 GnomeVFSFileSize *bytes,
+			 gchar **mime_type, 
+			 GError **error)
 {
 	GnomeVFSFileInfo *info;
-	GnomeVFSFileSize bytes;
 	GnomeVFSResult result;
 
 	info = gnome_vfs_file_info_new ();
@@ -500,27 +563,33 @@ eog_image_determine_file_bytes (EogImage *img, GError **error)
 	result = gnome_vfs_get_file_info_uri (img->priv->uri,
 					      info,
 					      GNOME_VFS_FILE_INFO_DEFAULT |
+					      GNOME_VFS_FILE_INFO_GET_MIME_TYPE |
 					      GNOME_VFS_FILE_INFO_FOLLOW_LINKS);
 
 	if ((result != GNOME_VFS_OK) || 
-	    (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_SIZE) == 0) {
-		bytes = 0;
+	    (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_SIZE) == 0 ||
+	    (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE) == 0) {
+		if (bytes)
+			*bytes = 0;
+
+		if (mime_type)
+			*mime_type = NULL;
 
 		g_set_error (error, 
 			     EOG_IMAGE_ERROR, 
 			     EOG_IMAGE_ERROR_VFS,
 			     gnome_vfs_result_to_string (result));
 	} else {
-		bytes = info->size;
+		if (bytes)
+			*bytes = info->size;
+
+		if (mime_type)
+			*mime_type = g_strdup (info->mime_type);
 	}
 
 	gnome_vfs_file_info_unref (info);
-
-	return bytes;
 }
 
-
-
 #ifdef HAVE_LCMS
 void
 eog_image_apply_display_profile (EogImage *img, cmsHPROFILE screen)
@@ -744,15 +813,7 @@ eog_image_real_autorotate (EogImage *img)
 		lookup[priv->orientation - 1] : EOG_TRANSFORM_NONE);
 
 	if (type != EOG_TRANSFORM_NONE) {
-		EogTransform *trans = eog_transform_new (type);
-
-		if (priv->trans == NULL) {
-			priv->trans = g_object_ref (trans);
-		} else {
-			eog_transform_compose (priv->trans, trans);
-		}
-		
-		g_object_unref (trans);
+		img->priv->trans_autorotate = eog_transform_new (type);
 	}
 
 	/* Disable auto orientation for next loads */
@@ -798,16 +859,19 @@ eog_image_set_exif_data (EogImage *img, EogMetadataReader *md_reader)
 	priv->exif_chunk = NULL;
 	priv->exif_chunk_len = 0;
 
-	/* EXIF data is already available, set the image
-           orientation and emit size prepared signal */
+	/* EXIF data is already available, set the image orientation */
 	if (priv->autorotate) {
 		eog_image_set_orientation (img);
 
-		g_signal_emit (img, 
-			       signals[SIGNAL_SIZE_PREPARED], 
-			       0, 
-			       priv->width, 
-			       priv->height);
+		/* Emit size prepared signal if we have the size */
+		if (priv->width > 0 &&
+		    priv->height > 0) {
+			g_signal_emit (img, 
+				       signals[SIGNAL_SIZE_PREPARED], 
+				       0, 
+				       priv->width, 
+				       priv->height);
+		}
 	}
 #else
 	eog_metadata_reader_get_exif_chunk (md_reader, 
@@ -820,7 +884,6 @@ eog_image_set_exif_data (EogImage *img, EogMetadataReader *md_reader)
  * Attempts to get the image dimensions from the thumbnail.
  * Returns FALSE if this information is not found.
  **/
-
 static gboolean
 eog_image_get_dimension_from_thumbnail (EogImage *image,
 			                gint     *width,
@@ -859,6 +922,7 @@ eog_image_real_load (EogImage *img,
 	GnomeVFSResult result;
 	EogMetadataReader *md_reader = NULL;
 	GdkPixbufFormat *format;
+	gchar *mime_type;
 	GdkPixbufLoader *loader = NULL;
 	guchar *buffer;
 	gboolean failed = FALSE;
@@ -877,9 +941,12 @@ eog_image_real_load (EogImage *img,
 		priv->file_type = NULL;
 	}
 
-	priv->bytes = eog_image_determine_file_bytes (img, error);
+	priv->threadsafe_format = FALSE;
 
-	if (priv->bytes == 0) {
+	eog_image_get_file_info (img, &priv->bytes, &mime_type, error);
+
+	if (*error) {
+		g_free (mime_type);
 		return FALSE;
 	}
 
@@ -902,6 +969,8 @@ eog_image_real_load (EogImage *img,
 	result = gnome_vfs_open_uri (&handle, priv->uri, GNOME_VFS_OPEN_READ);
 
 	if (result != GNOME_VFS_OK) {
+		g_free (mime_type);
+
 		g_set_error (error, 
 			     EOG_IMAGE_ERROR, 
 			     EOG_IMAGE_ERROR_VFS,
@@ -911,9 +980,24 @@ eog_image_real_load (EogImage *img,
 	}
 	
 	buffer = g_new0 (guchar, EOG_IMAGE_READ_BUFFER_SIZE);
-
+	
 	if (read_image_data || read_only_dimension) {
-		loader = gdk_pixbuf_loader_new ();
+		loader = gdk_pixbuf_loader_new_with_mime_type (mime_type, error);
+	
+		if (*error) {
+			g_error_free (*error);
+			*error = NULL;
+
+			loader = gdk_pixbuf_loader_new (); 
+		}
+
+		/* This is used to detect non-threadsafe loaders and disable
+ 		 * any possible asyncronous task that could bring deadlocks
+ 		 * to image loading process. */
+		g_signal_connect (loader, 
+				  "size-prepared", 
+				  G_CALLBACK (eog_image_pre_size_prepared),
+				  img);
 
 		g_signal_connect_object (G_OBJECT (loader), 
 					 "size-prepared", 
@@ -921,6 +1005,7 @@ eog_image_real_load (EogImage *img,
 					 img, 
 					 0);
         }
+	g_free (mime_type);
 
 	while (!priv->cancel_loading) {
 		result = gnome_vfs_read (handle, 
@@ -966,11 +1051,12 @@ eog_image_real_load (EogImage *img,
 					break;
 				}
 
-				g_signal_emit (img, 
-					       signals[SIGNAL_SIZE_PREPARED], 
-					       0, 
-					       priv->width, 
-					       priv->height);
+				if (priv->threadsafe_format)
+					g_signal_emit (img, 
+						       signals[SIGNAL_SIZE_PREPARED], 
+						       0, 
+						       priv->width, 
+						       priv->height);
                         }
 
 			first_run = FALSE;
@@ -1005,11 +1091,9 @@ eog_image_real_load (EogImage *img,
 	}
 
 	if (read_image_data || read_only_dimension) {
-		if (failed) {
-			gdk_pixbuf_loader_close (loader, NULL);
-		} else if (!gdk_pixbuf_loader_close (loader, error)) {
-			failed = TRUE;
-		}
+		/* Close silently in order to support partial
+		 * images as well. */
+		gdk_pixbuf_loader_close (loader, NULL);
         }
 
 	g_free (buffer);
@@ -1019,8 +1103,7 @@ eog_image_real_load (EogImage *img,
 	failed = (failed ||
 		  priv->cancel_loading || 
 		  bytes_read_total == 0 || 
-		  *error != NULL || 
-		  priv->cancel_loading);
+		  *error != NULL);
 
 	if (failed) {
 		if (priv->cancel_loading) {
@@ -1049,6 +1132,15 @@ eog_image_real_load (EogImage *img,
 			if (format != NULL) {
 				priv->file_type = gdk_pixbuf_format_get_name (format);
 			}
+
+			/* If it's non-threadsafe loader, then trigger window 
+ 			 * showing in the end of the process. */
+			if (!priv->threadsafe_format)
+				g_signal_emit (img, 
+					       signals[SIGNAL_SIZE_PREPARED], 
+					       0, 
+					       priv->width, 
+					       priv->height);
 		}
 	}
 
@@ -1425,6 +1517,11 @@ eog_image_reset_modifications (EogImage *image)
 		priv->trans = NULL;
 	}
 
+	if (priv->trans_autorotate != NULL) {
+		g_object_unref (priv->trans_autorotate);
+		priv->trans_autorotate = NULL;
+	}
+
 	priv->modified = FALSE;
 }
 
@@ -1714,9 +1811,6 @@ eog_image_get_caption (EogImage *img)
 		/* Use cached caption string */
 		return priv->caption;
 	
-	/* g_print ("priv->uri: %p\n", priv->uri);
-	   g_print ("%s\n", gnome_vfs_uri_to_string (priv->uri, GNOME_VFS_URI_HIDE_NONE)); */
-
 	name = gnome_vfs_uri_extract_short_name (priv->uri);
 	
 	if (name != NULL && g_ascii_strcasecmp (priv->uri->method_string, "file") == 0) {
